from typing import Type, Tuple

from pydantic import BaseModel

from pyappconf.model import BaseConfig
from tests.config import JSON_PATH, YAML_PATH, TOML_PATH
from tests.fixtures.model import model_object, model_classes


def test_to_json(model_object: BaseConfig):
    assert model_object.to_json(json_kwargs=dict(indent=2)) == JSON_PATH.read_text()


def test_to_yaml(model_object: BaseConfig):
    assert model_object.to_yaml() == YAML_PATH.read_text()


def test_to_toml(model_object: BaseConfig):
    assert model_object.to_toml() == TOML_PATH.read_text()


def test_from_json(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, SubConfig = model_classes
    loaded_object = MyConfig.parse_json(JSON_PATH)
    assert model_object == loaded_object


def test_from_yaml(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, SubConfig = model_classes
    loaded_object = MyConfig.parse_yaml(YAML_PATH)
    assert model_object == loaded_object


def test_from_toml(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, SubConfig = model_classes
    loaded_object = MyConfig.parse_toml(TOML_PATH)
    assert model_object == loaded_object


from enum import Enum
from pathlib import Path, PosixPath

from yaml.emitter import Emitter
from yaml.representer import SafeRepresenter
from yaml.resolver import Resolver
from yaml.serializer import Serializer

from pyappconf.encoding.general import HasStr


def _represent_hasstr(obj, value: HasStr):
    from yaml.representer import SafeRepresenter

    return SafeRepresenter.represent_str(obj, str(value))


def _represent_enum(obj, value: Enum):
    from yaml.representer import SafeRepresenter

    return SafeRepresenter.represent_str(obj, value.value)


class CustomRepresenter(SafeRepresenter):
    pass


CustomRepresenter.add_representer(Path, _represent_hasstr)
CustomRepresenter.add_representer(PosixPath, _represent_hasstr)
CustomRepresenter.add_multi_representer(Enum, _represent_enum)


class CustomDumper(Emitter, Serializer, CustomRepresenter, Resolver):
    def __init__(
        self,
        stream,
        default_style=None,
        default_flow_style=False,
        canonical=None,
        indent=None,
        width=None,
        allow_unicode=None,
        line_break=None,
        encoding=None,
        explicit_start=None,
        explicit_end=None,
        version=None,
        tags=None,
        sort_keys=True,
    ):
        Emitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            allow_unicode=allow_unicode,
            line_break=line_break,
        )
        Serializer.__init__(
            self,
            encoding=encoding,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags,
        )
        CustomRepresenter.__init__(
            self,
            default_style=default_style,
            default_flow_style=default_flow_style,
            sort_keys=sort_keys,
        )
        Resolver.__init__(self)



from enum import Enum
from pathlib import Path, PosixPath
from typing import Dict, Callable, Type, Any

from toml.encoder import _dump_str, TomlEncoder

from pyappconf.encoding.general import HasStr


def _dump_hasstr(obj: HasStr) -> str:
    return _dump_str(str(obj))


def _dump_enum(obj: Enum) -> str:
    return _dump_str(obj.value)


class CustomTomlEncoder(TomlEncoder):
    multi_dump_funcs: Dict[Type, Callable[[Any], str]]

    def __init__(self, _dict=dict, preserve=False):
        super().__init__(_dict=_dict, preserve=preserve)
        self.multi_dump_funcs = {Enum: _dump_enum}
        self.dump_funcs[Path] = _dump_hasstr
        self.dump_funcs[PosixPath] = _dump_hasstr

    def dump_value(self, v):
        """
        Overrides TomlEncoder.dump_value to adds support for putting
        a base class in multi_dump_funcs and have it be used for any
        subclass of that type
        """
        # Lookup function corresponding to v's type
        dump_fn = self.dump_funcs.get(type(v))
        if dump_fn is None:
            for base_cls, fn in self.multi_dump_funcs.items():
                if isinstance(v, base_cls):
                    dump_fn = fn
            if dump_fn is not None:
                # Got a multi dump function
                return dump_fn(v)
        return super().dump_value(v)



import os
from copy import deepcopy
from enum import Enum
from pathlib import Path
from typing import Optional, Union, Dict, Any, Callable, Type

from pydantic import BaseSettings, validator, BaseModel
import yaml
import toml
import json
import appdirs
from pydantic.env_settings import EnvSettingsSource
from toml.encoder import TomlEncoder

from pyappconf.encoding.ext_json import ExtendedJSONEncoder
from pyappconf.encoding.ext_toml import CustomTomlEncoder
from pyappconf.encoding.ext_yaml import CustomDumper


def _output_if_necessary(content: str, out_path: Optional[Union[str, Path]] = None):
    if out_path is not None:
        out_path = Path(out_path)
        out_path.write_text(content)


def _get_data_kwargs(**kwargs):
    default_kwargs = dict(
        exclude={"settings"},
    )
    if "exclude" in kwargs:
        default_kwargs["exclude"].update(kwargs["exclude"])
    kwargs.update(default_kwargs)
    return kwargs


class ConfigFormats(str, Enum):
    YAML = "yaml"
    JSON = "json"
    TOML = "toml"

    @classmethod
    def from_path(cls, path: Path) -> "ConfigFormats":
        ext = path.suffix.strip(".").casefold()
        if ext in ("yml", "yaml"):
            return cls.YAML
        if ext == "json":
            return cls.JSON
        if ext == "toml":
            return cls.TOML
        raise ValueError(
            f"suffix {ext} not a supported config format. Supplied path: {path}"
        )


class AppConfig(BaseModel):
    app_name: str
    config_name: str = "config"
    custom_config_folder: Optional[Path] = None
    default_format: ConfigFormats = ConfigFormats.TOML
    toml_encoder: Type[TomlEncoder] = CustomTomlEncoder
    yaml_encoder: Type = CustomDumper
    json_encoder: Type[json.JSONEncoder] = ExtendedJSONEncoder

    @property
    def config_base_location(self) -> Path:
        if self.custom_config_folder is not None:
            return self.custom_config_folder / self.config_name
        return Path(appdirs.user_config_dir(self.app_name)) / self.config_name

    @property
    def config_location(self) -> Path:
        return Path(str(self.config_base_location) + "." + self.default_format.value)

    @property
    def config_file_name(self) -> str:
        return self.config_location.name


class BaseConfig(BaseSettings):
    _settings: AppConfig
    settings: AppConfig = None  # type: ignore

    @validator("settings")
    def set_settings_from_class_if_none(cls, v):
        if v is None:
            return cls._settings
        return v

    def get_serializer(
        self,
    ) -> Callable[[Optional[Union[str, Path]], Optional[Dict[str, Any]]], str]:
        if self.settings.default_format == ConfigFormats.TOML:
            return self.to_toml
        if self.settings.default_format == ConfigFormats.YAML:
            return self.to_yaml
        if self.settings.default_format == ConfigFormats.JSON:
            return self.to_json
        raise NotImplementedError(f"unsupported format {self.settings.default_format}")

    @classmethod
    def get_deserializer(
        cls, config_format: Optional[ConfigFormats] = None
    ) -> Callable[[Union[str, Path]], "BaseConfig"]:
        if config_format is None:
            config_format = cls._settings.default_format

        if config_format == ConfigFormats.TOML:
            return cls.parse_toml
        if config_format == ConfigFormats.YAML:
            return cls.parse_yaml
        if config_format == ConfigFormats.JSON:
            return cls.parse_json
        raise NotImplementedError(f"unsupported format {config_format}")

    @classmethod
    def _settings_with_overrides(cls, **kwargs) -> AppConfig:
        if not kwargs:
            return deepcopy(cls._settings)
        config_data = cls._settings.dict()
        config_data.update(**kwargs)
        return AppConfig(**config_data)

    def save(self, serializer_kwargs: Optional[Dict[str, Any]] = None, **kwargs):
        if not self.settings.config_location.parent.exists():
            self.settings.config_location.parent.mkdir(parents=True)
        self.get_serializer()(self.settings.config_location, serializer_kwargs, **kwargs)  # type: ignore

    @classmethod
    def load(cls, path: Optional[Union[str, Path]] = None) -> "BaseConfig":
        assign_settings = True

        config_format: Optional[ConfigFormats]
        if path is None:
            assign_settings = False
            path = cls._settings.config_location
            config_format = None
        else:
            path = Path(path)
            config_format = ConfigFormats.from_path(path)

        obj = cls.get_deserializer(config_format)(path)
        if assign_settings:
            obj.settings = cls._settings_with_overrides(
                custom_config_folder=path.parent,
                default_format=config_format,
                config_name=path.stem,
            )
        return obj

    @classmethod
    def load_or_create(cls, path: Optional[Union[str, Path]] = None) -> "BaseConfig":
        if path is None:
            path = cls._settings.config_location
        else:
            path = Path(path)
        if path.exists():
            return cls.load(path)
        else:
            config_format = ConfigFormats.from_path(path)
            return cls(
                settings=cls._settings_with_overrides(
                    custom_config_folder=path.parent,
                    default_format=config_format,
                    config_name=path.stem,
                )
            )

    @classmethod
    def load_recursive(cls, path: Optional[Union[str, Path]] = None) -> "BaseConfig":
        """
        Searches the passed path or current directory for a config with the correct name,
        and if not found goes the parent directory and repeats the search.
        If the config is not found after reaching the root directory,
        it will look at the location in the config.

        :param path: The path to start searching from, defaults to the current directory
        :return:
        """
        path = Path(path or os.getcwd()).absolute()
        current_path = path

        def has_hit_root_directory() -> bool:
            return current_path.parent == current_path

        file_name = cls._settings.config_file_name
        while True:
            check_path = current_path / file_name
            if check_path.exists():
                return cls.load(check_path)
            if has_hit_root_directory():
                break
            current_path = current_path.parent

        # Could not find config after reaching root directory. Try
        # loading from default location
        return cls.load()

    @classmethod
    def _get_env_values(cls) -> Dict[str, Any]:
        env_file = getattr(cls.Config, "env_file", None)
        source = EnvSettingsSource(env_file=env_file, env_file_encoding=None)
        return source(cls)  # type: ignore

    def to_yaml(
        self,
        out_path: Optional[Union[str, Path]] = None,
        yaml_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> str:
        if yaml_kwargs is None:
            yaml_kwargs = {}
        kwargs = _get_data_kwargs(**kwargs)
        data = self.dict(**kwargs)
        yaml_str = yaml.dump(data, **yaml_kwargs, Dumper=self.settings.yaml_encoder)
        _output_if_necessary(yaml_str, out_path)
        return yaml_str

    @classmethod
    def parse_yaml(cls, in_path: Union[str, Path]) -> "BaseConfig":
        data = yaml.safe_load(Path(in_path).read_text())
        data.update(cls._get_env_values())
        return cls(**data)

    def to_toml(
        self,
        out_path: Optional[Union[str, Path]] = None,
        toml_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> str:
        if toml_kwargs is None:
            toml_kwargs = {}
        kwargs = _get_data_kwargs(**kwargs)
        data = self.dict(**kwargs)
        toml_str = toml.dumps(data, **toml_kwargs, encoder=self.settings.toml_encoder())  # type: ignore
        _output_if_necessary(toml_str, out_path)
        return toml_str

    @classmethod
    def parse_toml(cls, in_path: Union[str, Path]) -> "BaseConfig":
        data = toml.load(in_path)
        data.update(cls._get_env_values())
        return cls(**data)

    def to_json(
        self,
        out_path: Optional[Union[str, Path]] = None,
        json_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> str:
        if json_kwargs is None:
            json_kwargs = {}
        kwargs = _get_data_kwargs(**kwargs)
        data = self.dict(**kwargs)
        json_str = json.dumps(data, **json_kwargs, cls=self.settings.json_encoder)
        _output_if_necessary(json_str, out_path)
        return json_str

    @classmethod
    def parse_json(cls, in_path: Union[str, Path]) -> "BaseConfig":
        data = json.loads(Path(in_path).read_text())
        data.update(cls._get_env_values())
        return cls(**data)



from pathlib import Path

TEST_DIR = Path(__file__).parent
INPUT_DATA_DIR = TEST_DIR / 'input_data'
GENERATED_DATA_DIR = TEST_DIR / 'generated_data'

DATA_NAME = 'data'
JSON_PATH = INPUT_DATA_DIR / (DATA_NAME + '.json')
YAML_PATH = INPUT_DATA_DIR / (DATA_NAME + '.yaml')
TOML_PATH = INPUT_DATA_DIR / (DATA_NAME + '.toml')
NON_EXISTENT_NAME = 'woo'
NON_EXISTENT_IN_PATH = INPUT_DATA_DIR / NON_EXISTENT_NAME
NON_EXISTENT_INPUT_JSON_PATH = INPUT_DATA_DIR / (NON_EXISTENT_NAME + '.json')
CUSTOM_CONFIG_IN_PATH = INPUT_DATA_DIR / DATA_NAME
ENV_PATH = GENERATED_DATA_DIR / '.env'
RECURSIVE_INPUT_FOLDER = INPUT_DATA_DIR / "recursive"

if not INPUT_DATA_DIR.exists():
    INPUT_DATA_DIR.mkdir()
if not GENERATED_DATA_DIR.exists():
    GENERATED_DATA_DIR.mkdir()


from enum import Enum
from pathlib import Path
from typing import Type, Dict, List, Tuple, Sequence, Optional

import pytest
from pydantic import BaseModel, Field

from pyappconf.model import BaseConfig, AppConfig
from tests.config import ENV_PATH


def get_model_classes() -> Tuple[Type[BaseConfig], Type[BaseModel]]:
    class MyEnum(str, Enum):
        ONE = "one"
        TWO = "two"

    class SubConfig(BaseModel):
        a: str
        b: float

    class MyConfig(BaseConfig):
        string: str
        integer: int
        custom: SubConfig
        dictionary: Dict[str, SubConfig]
        str_list: List[str]
        int_tuple: Tuple[int, ...]

        default_string: str = "woo"
        default_custom: SubConfig = SubConfig(a="yeah", b=5.6)
        default_enum: MyEnum = MyEnum.ONE
        default_enum_list: List[MyEnum] = Field(default_factory=lambda: [MyEnum.ONE, MyEnum.TWO])
        file_path: Path = Path('/a/b.txt')

        _settings: AppConfig = AppConfig(app_name="MyApp")

        class Config:
            env_prefix = "MYAPP_"
            env_file = ENV_PATH

    return MyConfig, SubConfig


def get_model_class_with_defaults() -> Type[BaseConfig]:
    class MyDefaultConfig(BaseConfig):
        a: str = "a"
        b: int = 2

        _settings: AppConfig = AppConfig(app_name="MyApp")

        class Config:
            env_prefix = "MYAPP_"
            env_file = ENV_PATH

    return MyDefaultConfig


def get_model_object(
    exclude_keys: Optional[Sequence[str]] = None, **kwargs
) -> BaseConfig:
    MyConfig, SubConfig = get_model_classes()

    all_kwargs = dict(
        string="a",
        integer=5,
        custom=SubConfig(a="b", b=8.5),
        dictionary={"yeah": SubConfig(a="c", b=9.6)},
        str_list=["a", "b", "c"],
        int_tuple=(1, 2, 3),
    )
    if exclude_keys is not None:
        all_kwargs = {
            key: value for key, value in all_kwargs.items() if key not in exclude_keys
        }
    all_kwargs.update(kwargs)

    # conf = MyConfig(**{'myapp_' + key: value for key, value in all_kwargs.items()})
    conf = MyConfig(**all_kwargs)
    return conf


def get_model_object_with_defaults() -> BaseConfig:
    MyDefaultConfig = get_model_class_with_defaults()
    return MyDefaultConfig()


@pytest.fixture(scope="session")
def model_classes() -> Tuple[Type[BaseConfig], Type[BaseModel]]:
    return get_model_classes()


@pytest.fixture(scope="session")
def model_class_with_defaults() -> Type[BaseConfig]:
    return get_model_class_with_defaults()


@pytest.fixture(scope="session")
def model_object() -> BaseConfig:
    return get_model_object()


@pytest.fixture(scope="session")
def model_object_with_defaults() -> BaseConfig:
    return get_model_object_with_defaults()



import shutil
from pathlib import Path
from typing import Type, Tuple, Optional

import pytest
from pydantic import BaseModel

from pyappconf.model import AppConfig, ConfigFormats, BaseConfig
from tests.config import (
    GENERATED_DATA_DIR,
    JSON_PATH,
    YAML_PATH,
    TOML_PATH,
    DATA_NAME,
    NON_EXISTENT_INPUT_JSON_PATH,
    NON_EXISTENT_NAME,
    RECURSIVE_INPUT_FOLDER,
    INPUT_DATA_DIR,
)
from tests.fixtures.model import (
    get_model_object,
    get_model_classes,
    model_classes,
    model_object,
    model_class_with_defaults,
    model_object_with_defaults,
)


def _save_load_test(custom_settings: AppConfig):
    mod = get_model_object(settings=custom_settings)
    mod.save()
    assert str(mod.settings.config_location).endswith(
        custom_settings.default_format.value
    )

    OrigConfig, SubConfig = get_model_classes()

    class MyConfig(OrigConfig):
        _settings = custom_settings

    assert str(MyConfig._settings.config_location).endswith(
        custom_settings.default_format.value
    )
    obj = MyConfig.load()
    assert mod == obj


def test_save_load_toml():
    custom_settings = AppConfig(
        app_name="MyApp", custom_config_folder=GENERATED_DATA_DIR
    )
    _save_load_test(custom_settings)


def test_save_load_yaml():
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=GENERATED_DATA_DIR,
        default_format=ConfigFormats.YAML,
    )
    _save_load_test(custom_settings)


def test_save_load_json():
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=GENERATED_DATA_DIR,
        default_format=ConfigFormats.JSON,
    )
    _save_load_test(custom_settings)


def assert_model_loaded_with_extension(
    mod: BaseConfig, model_object: BaseConfig, config_format: ConfigFormats
):
    assert mod.settings.custom_config_folder == INPUT_DATA_DIR
    assert mod.settings.default_format == config_format
    assert mod.settings.config_name == DATA_NAME

    mod.settings = model_object.settings
    assert mod == model_object


def test_load_toml_with_custom_path(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, _ = model_classes
    mod = MyConfig.load(TOML_PATH)
    assert_model_loaded_with_extension(mod, model_object, ConfigFormats.TOML)


def test_load_yaml_with_custom_path(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, _ = model_classes
    mod = MyConfig.load(YAML_PATH)
    assert_model_loaded_with_extension(mod, model_object, ConfigFormats.YAML)


def test_load_json_with_custom_path(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, _ = model_classes
    mod = MyConfig.load(JSON_PATH)
    assert_model_loaded_with_extension(mod, model_object, ConfigFormats.JSON)


def test_load_or_create_with_path_exists(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, _ = model_classes
    mod = MyConfig.load_or_create(JSON_PATH)
    assert_model_loaded_with_extension(mod, model_object, ConfigFormats.JSON)


def test_load_or_create_with_path_does_not_exist(
    model_object_with_defaults: BaseConfig, model_class_with_defaults: Type[BaseConfig]
):
    MyConfig = model_class_with_defaults
    model_object = model_object_with_defaults
    mod = MyConfig.load_or_create(NON_EXISTENT_INPUT_JSON_PATH)
    assert mod.settings.custom_config_folder == INPUT_DATA_DIR
    assert mod.settings.default_format == ConfigFormats.JSON
    assert mod.settings.config_name == NON_EXISTENT_NAME

    mod.settings = model_object.settings
    assert mod == model_object


@pytest.mark.parametrize(
    "path, expect_string",
    [
        (RECURSIVE_INPUT_FOLDER / "1", "loaded from 1"),
        (RECURSIVE_INPUT_FOLDER / "2", "loaded from recursive"),
        # Third and fourth cases - didn't find recursive, go to default location
        (INPUT_DATA_DIR, "a"),
        (None, "a"),
    ],
)
def test_load_recursive(path: Optional[Path], expect_string: str):
    OrigConfig, SubConfig = get_model_classes()

    shutil.copy(
        INPUT_DATA_DIR / "data.toml", GENERATED_DATA_DIR / "recursive-config.toml"
    )

    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=GENERATED_DATA_DIR,
        config_name="recursive-config",
    )

    class MyConfig(OrigConfig):
        _settings = custom_settings

    mod = MyConfig.load_recursive(path)

    assert mod.string == expect_string



"""
Generate test data for file output
"""
from tests.config import INPUT_DATA_DIR, DATA_NAME, JSON_PATH, YAML_PATH, TOML_PATH
from tests.fixtures.model import get_model_object

if __name__ == "__main__":
    conf = get_model_object()
    conf.to_json(out_path=JSON_PATH, json_kwargs=dict(indent=2))
    conf.to_yaml(out_path=YAML_PATH)
    conf.to_toml(out_path=TOML_PATH)



from typing import Tuple, Type
import os

from pyappconf import BaseConfig
from tests.config import TOML_PATH, ENV_PATH
from tests.fixtures.model import get_model_object, model_classes


def test_override_from_env(
    model_classes: Tuple[Type[BaseConfig], Type[BaseConfig]]
):
    MyConfig, SubConfig = model_classes
    os.environ['MYAPP_STRING'] = 'abc'
    os.environ['MYAPP_INT_TUPLE'] = '[4, 5, 6]'

    model_object = get_model_object(string='abc', int_tuple=(4, 5, 6))

    loaded_object = MyConfig.parse_toml(TOML_PATH)

    del os.environ['MYAPP_STRING']
    del os.environ['MYAPP_INT_TUPLE']
    assert model_object == loaded_object


def test_override_from_dotenv(
        model_classes: Tuple[Type[BaseConfig], Type[BaseConfig]]
):
    MyConfig, SubConfig = model_classes
    ENV_PATH.write_text("""
MYAPP_STRING=abc
MYAPP_INT_TUPLE="[4, 5, 6]"
    """)

    model_object = get_model_object(string='abc', int_tuple=(4, 5, 6))

    loaded_object = MyConfig.parse_toml(TOML_PATH)

    os.remove(ENV_PATH)
    assert model_object == loaded_object


from pathlib import Path
from unittest.mock import patch

from pyappconf.model import BaseConfig, AppConfig
from tests.fixtures.model import model_object, get_model_object


@patch("sys.platform", "linux")
def test_config_base_location_linux(model_object: BaseConfig):
    assert (
        model_object.settings.config_base_location
        == Path("~").expanduser() / ".config" / "MyApp" / "config"
    )


@patch("sys.platform", "linux")
def test_config_location_linux(model_object: BaseConfig):
    assert (
        model_object.settings.config_location
        == Path("~").expanduser() / ".config" / "MyApp" / "config.toml"
    )


def test_custom_config_folder():
    folder = Path("/woo")
    obj = get_model_object(
        settings=AppConfig(app_name="MyApp", custom_config_folder=folder)
    )
    assert obj.settings.config_base_location == (folder / "config")


def test_custom_config_name():
    folder = Path("/woo")
    obj = get_model_object(
        settings=AppConfig(app_name="MyApp", custom_config_folder=folder, config_name="yeah")
    )
    assert obj.settings.config_base_location == (folder / "yeah")


import json
from pathlib import Path
from typing import Any


class ExtendedJSONEncoder(json.JSONEncoder):

    def default(self, o: Any) -> Any:
        if isinstance(o, Path):
            return str(o)

        return json.JSONEncoder.default(self, o)